// $Id: filter-clamav.go 42 2019-11-10 20:34:32Z umaxx $
// Copyright (c) 2019 Joerg Jung <mail@umaxx.net>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

// filter-clamav - opensmtpd filter for clamav

package main

import (
	"bufio"
	"encoding/binary"
	"fmt"
	"log"
	"log/syslog"
	"net"
	"os"
	"strconv"
	"strings"
)

const (
	v  = "0.1"
	yr = "2019"
)

type clamav struct {
	sid string
	buf []string
	sz  uint64
	r   *bool
}

var (
	addr        = "localhost:3310"
	lim  uint64 = 26214400 // size limit in bytes
	l3   *syslog.Writer
	cls  = make(map[string]*clamav)
)

func (cl *clamav) reset(tok string) {
	if cl.buf != nil {
		for _, v := range cl.buf {
			fmt.Printf("filter-dataline|%s|%s|%s\n", tok, cl.sid, v)
		}
	}
	cl.buf = nil
	cl.sz = 0
	cl.r = nil
}

func (cl *clamav) status(tok string, ln string) {
	var s string
	l3.Info(fmt.Sprintln(cl.sid, "result", ln))
	if _, e := fmt.Sscanf(ln, "stream: %s", &s); e != nil {
		l3.Err(fmt.Sprintln(cl.sid, "sscanf", e))
		cl.reset(tok)
		return
	}
	cl.r = new(bool)
	*cl.r = (s != "OK")
}

func (cl *clamav) response(tok string, in *bufio.Scanner) {
	for in.Scan() {
		cl.status(tok, in.Text())
	}
	if e := in.Err(); e != nil {
		l3.Err(fmt.Sprintln(cl.sid, "scanner", e))
		cl.reset(tok)
		return
	}
}

func (cl *clamav) process(tok string) {
	con, e := net.Dial("tcp", addr)
	if e != nil {
		l3.Err(fmt.Sprintln(cl.sid, e))
		cl.reset(tok)
		return
	}
	defer con.Close()
	if _, e = fmt.Fprintf(con, "nINSTREAM\n"); e != nil {
		l3.Err(fmt.Sprintln(cl.sid, "write", e))
		cl.reset(tok)
		return
	}
	b := make([]byte, 4)
	for _, v := range cl.buf {
		binary.BigEndian.PutUint32(b, uint32(len(v)+1))
		if _, e = fmt.Fprintf(con, "%s%s\n", b, v); e != nil {
			l3.Err(fmt.Sprintln(cl.sid, "write", e))
			cl.reset(tok)
			return
		}
	}
	binary.BigEndian.PutUint32(b, 0)
	if _, e = fmt.Fprintf(con, "%s", b); e != nil {
		l3.Err(fmt.Sprintln(cl.sid, "write", e))
		cl.reset(tok)
		return
	}
	if c, ok := con.(*net.TCPConn); ok {
		if e = c.CloseWrite(); e != nil {
			l3.Warning(fmt.Sprintln(cl.sid, "closewrite", e))
		}
	}
	cl.response(tok, bufio.NewScanner(con))
	if cl.r != nil {
		for _, v := range cl.buf {
			fmt.Printf("filter-dataline|%s|%s|%s\n", tok, cl.sid, v)
		}
	}
	fmt.Printf("filter-dataline|%s|%s|.\n", tok, cl.sid)
}

func (cl *clamav) skip(tok string) {
	cl.reset(tok)
	cl.r = new(bool)
	*cl.r = false
}

func (cl *clamav) line(tok string, ln string) {
	if cl.sz > lim+1 { // +1 don't count eom dot '.' into limit
		l3.Warning(fmt.Sprintln(cl.sid, "limit reached, skip scan"))
		cl.skip(tok)
	}
	if cl.r != nil && *cl.r == false { // skipped message
		fmt.Printf("filter-dataline|%s|%s|%s\n", tok, cl.sid, ln)
		return
	}
	if ln == "." {
		go cl.process(tok)
		return
	}
	cl.buf = append(cl.buf, ln)
	cl.sz += uint64(len(ln))
}

func (cl *clamav) commit(tok string) {
	if cl.r == nil {
		l3.Warning(fmt.Sprintln(cl.sid, "reject filter failed"))
		fmt.Printf("filter-result|%s|%s|reject|451 4.7.1 Virus filter failed\n", tok, cl.sid)
	} else if *cl.r {
		l3.Info(fmt.Sprintln(cl.sid, "reject virus"))
		fmt.Printf("filter-result|%s|%s|reject|554 5.7.1 Virus found\n", tok, cl.sid)
	} else {
		l3.Debug(fmt.Sprintln(cl.sid, "accept"))
		fmt.Printf("filter-result|%s|%s|proceed\n", tok, cl.sid)
	}
}

func register(in *bufio.Scanner) error {
	l3.Info("register")
	for in.Scan() { // skip config
		if in.Text() == "config|ready" {
			fmt.Println("register|report|smtp-in|link-connect")
			fmt.Println("register|filter|smtp-in|data-line")
			fmt.Println("register|filter|smtp-in|commit")
			fmt.Println("register|report|smtp-in|link-disconnect")
			fmt.Println("register|ready")
			return nil
		}
	}
	return in.Err()
}

func run() {
	l3.Info("start")
	defer l3.Info("exit")
	in := bufio.NewScanner(os.Stdin)
	if e := register(in); e != nil {
		l3.Err(fmt.Sprintln("register", e))
		return
	}
	for in.Scan() {
		f := strings.Split(in.Text(), "|")
		t, ver, ev, sid := f[0], f[1], f[4], f[5]
		if (t != "filter" && t != "report") || ver != "0.4" {
			l3.Err(fmt.Sprintln(sid, "protocol", t, ver))
			return
		}
		switch ev {
		case "link-connect":
			cls[sid] = &clamav{sid: sid, buf: nil, sz: 0, r: nil}
		case "data-line":
			if c, ok := cls[sid]; ok {
				c.line(f[6], strings.Join(f[7:], "|"))
			}
		case "commit":
			if c, ok := cls[sid]; ok {
				c.commit(f[6])
			}
		case "link-disconnect":
			if c, ok := cls[sid]; ok {
				delete(cls, c.sid)
			}
		default:
			l3.Err(fmt.Sprintln(sid, "event", ev))
			return
		}
	}
	if e := in.Err(); e != nil {
		l3.Err(fmt.Sprintln("scanner", e))
		return
	}
}

func init() {
	log.SetFlags(log.Lshortfile)
}

func main() {
	var e error
	if len(os.Args) == 2 && os.Args[1] == "version" {
		fmt.Println("filter-clamav", v, "(c)", yr, "Joerg Jung")
		return
	}
	if len(os.Args) > 3 {
		log.Fatalf("usage: filter-clamav [<address>] [<limit>]\n%35sfilter-clamav version\n", "")
	}
	if len(os.Args) >= 2 {
		addr = os.Args[1]
	}
	if len(os.Args) == 3 {
		if lim, e = strconv.ParseUint(os.Args[2], 10, 0); e != nil {
			log.Fatal(e)
		}
		if lim == 0 {
			log.Fatal("limit must be larger than 0")
		}
	}
	if l3, e = syslog.New(syslog.LOG_MAIL, "filter-clamav"); e != nil {
		log.Fatal(e)
	}
	defer l3.Close()
	run()
}
